from pathlib import PureWindowsPath, Path

WINDOWS_SANDBOX_DEFAULT_DESKTOP = Path(PureWindowsPath(r'C:\Users\WDAGUtilityAccount\Desktop'))


def shared_folder_path_in_sandbox(local_path):
    local_path = Path(PureWindowsPath(local_path))
    if local_path.exists() and not local_path.is_dir():
        local_path = local_path.parent

    return WINDOWS_SANDBOX_DEFAULT_DESKTOP / local_path.name



from .config.config_genereator import generate_config_file
from .session import offline_session, online_session

import copy
import subprocess
import cached_property
import rpyc
from warnings import warn_explicit


class OfflineSandbox:
    def __init__(self, config):
        self.config = config
        offline_session.OfflineSession(self).run()

    @staticmethod
    def start_sandbox(config_file_path):
        subprocess.Popen(['start', config_file_path],
                         shell=True)


class OnlineSandbox:
    def __init__(self, config, launch_new_instance=True):
        self.config = config

        assert config.networking, "Networking not configured with an online sandbox."
        if len(self.config.logon_script) != 0:
            warn_explicit("Logon scripts are ignored when the sandbox has networking enabled.")

        session = online_session.OnlineSession(self)

        # Try to connect to an already running server.
        self._connection_tuple = session.running_sandbox_server_information(launch_new_instance)

        # Server is down, let's boot the sandbox.
        if self._connection_tuple is None:
            session.configure_sandbox()
            OfflineSandbox(self.config)

        # And get the new sandbox connection tuple.
        self._connection_tuple = session.connect_to_sandbox()

    @cached_property.cached_property
    def rpyc(self):
        """
        RPyC connection to the sandbox.
        """

        assert self.config.networking, "Networking is not enabled in this Sandbox."

        if self._connection_tuple is not None:
            return rpyc.classic.connect(*self._connection_tuple)

    def run_executable(self, executable_args, *args, **kwargs):
        """
        Run an executable in the sandbox.
        :param executable_args: The executable arguments to run.
        :param args: Extra arguments to `subprocess.Popen`.
        :param kwargs: Extra kwargs to `subprocess.Popen`.
        :return: a remote `subprocess.Popen` instance.
        """

        kwargs['stdout'] = kwargs.pop('stdout', subprocess.PIPE)
        kwargs['stderr'] = kwargs.pop('stderr', subprocess.PIPE)
        return self.rpyc.modules.subprocess.Popen(executable_args, *args, **kwargs)

    def shutdown(self):
        """
        Shutdown the sandbox.
        """

        self.run_executable(['shutdown.exe', '/s', '/t', '0'])



import os
import time


def wait_for_file_creation(filename, timeout=25, check_interval=1):
    """
    Return true if filename exists, if not keep checking once every check_interval seconds for time_limit seconds.
    """

    if timeout < 1:
        timeout = 1

    now = time.time()
    timeout_time = now + timeout

    while time.time() <= timeout_time:
        if os.path.exists(filename):
            return True
        else:
            # Wait for check interval seconds, then check again.
            time.sleep(check_interval)

    return False



import sys
import pathlib
import site


class FolderMapper:
    """
    Generic Folder Mapper. Enables to map a folder to the new sandbox.
    """

    def __init__(self, folder_path, read_only=True):
        self._folder_path = pathlib.Path(folder_path)
        self._read_only = read_only

    def path(self):
        return self._folder_path

    def read_only(self):
        return self._read_only


class PythonMapper:
    """
    Maps the current Python installation to the new sandbox.
    """

    def path(self):
        return pathlib.Path(sys.prefix)

    def read_only(self):
        return True


class PythonUserSitePackagesMapper:
    """
    Maps the current Python installation's user site packages to the new sandbox.
    """

    def path(self):
        return pathlib.Path(site.getusersitepackages())

    def read_only(self):
        return True



"""
This script is the RPyC server that is uploaded to the Windows Sandbox.
It boots up the server, writes the IP:port data to a shared file and starts accepting connections.
"""

import argparse
from pathlib import Path
import socket
import sys
import subprocess
import site
import os

WINDOWS_SANDBOX_DEFAULT_DESKTOP = Path(r'C:\Users\WDAGUtilityAccount\Desktop')


def enable_python_incoming_firewall():
    # Using `{sys.base_exec_prefix}\python.exe` instead of `sys.executable` to support venvs.
    for python_executable in ['python.exe', 'python3.exe']:
        subprocess.run(
            'netsh advfirewall firewall add rule name=AllowPythonServer '
            'dir=in action=allow enable=yes program="{}\\{}"'.format(Path(sys.base_exec_prefix).resolve(), python_executable),
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )


def get_ip_address():
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("address_path",
                        type=lambda p: Path(p))
    parser.add_argument("custom_user_site_packages",
                        type=str,
                        default='',
                        nargs='?')
    parser.add_argument("--disable-firewall", '-f', default=False, action='store_true')
    args = parser.parse_args()

    if args.disable_firewall:
        enable_python_incoming_firewall()

    if args.custom_user_site_packages:
        site.addsitedir(args.custom_user_site_packages)

    import rpyc.utils.server

    server = rpyc.utils.server.ThreadedServer(rpyc.classic.ClassicService, port=0)

    Path(WINDOWS_SANDBOX_DEFAULT_DESKTOP / args.address_path).write_text('{}:{}'.format(get_ip_address(),
                                                                                        server.port))
    server.start()


if __name__ == '__main__':
    main()



import sys
import os
import argparse
import winreg
import pathlib

from IPython import embed
from traitlets.config import get_config

import winsandbox


def read_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--interactive", default=False, action="store_true",
                        help="Interactive mode with a `sandbox` instance available. Embeds an IPython shell.")
    parser.add_argument("--no-vgpu", default=False, action="store_true", help="Disable the virtual GPU.")
    parser.add_argument("-s", "--logon-script", required=False,
                        help="Logon script for non-interactive (ipy) sandbox.", default="")
    parser.add_argument("-f", "--folder", action="append",
                        help="Folders to map to the sandbox.", default=[])

    parser.add_argument("-r", "--register", default=False, action="store_true", 
                        help="Register a Shell extension that allows opening executable files in the sandbox")
    parser.add_argument("-u", "--unregister", default=False, action="store_true", 
                        help="Unregister the Shell extension")

    return parser.parse_args()


def register_shell_extension(name, cli_flags = ''):
    key = winreg.CreateKeyEx(winreg.HKEY_CLASSES_ROOT, r'exefile\shell\{}'.format(name))

    # Calculate paths
    package_root_dir = pathlib.Path(__file__).absolute().parent
    icon_path = package_root_dir / 'shell_extension' / 'sandbox.ico'

    # Set icon
    winreg.SetValueEx(key, 'Icon', 0, winreg.REG_SZ, str(icon_path))

    # Set shell script command
    command_key = winreg.CreateKeyEx(key, 'Command')
    command = '{} -m winsandbox.shell_extension.open_sandboxed "%1" {}'.format(sys.executable, cli_flags)
    winreg.SetValue(command_key, None, winreg.REG_SZ, command)

    print("Registered the '{}' shell extension successfully!".format(name))


def unregister_shell_extension(name):
    try:
        winreg.DeleteKey(winreg.HKEY_CLASSES_ROOT, r'exefile\shell\{}\Command'.format(name))
    except FileNotFoundError:
        print("Shell extension '{}' is not registered".format(name))
        return

    winreg.DeleteKey(winreg.HKEY_CLASSES_ROOT, r'exefile\shell\{}'.format(name))

    print("Unregistered the '{}' shell extension successfully!".format(name))


def main():

    args = read_args()

    # Handle shell extension commands
    try:
        if args.register:
            register_shell_extension('Open Sandboxed')
            register_shell_extension('Run Sandboxed', cli_flags='--run')
            return
        elif args.unregister:
            unregister_shell_extension('Open Sandboxed')
            unregister_shell_extension('Run Sandboxed')
            return
    except PermissionError:
        print("Try running again as an Administrator")
        return

    # Launch a new sandbox
    sandbox = winsandbox.new_sandbox(networking=args.interactive,
                                     virtual_gpu=not args.no_vgpu,
                                     logon_script=args.logon_script,
                                     folder_mappers=[winsandbox.FolderMapper(folder_path=folder, read_only=False)
                                                     for folder in args.folder])

    if args.interactive:
        config = get_config()
        config.InteractiveShellEmbed.colors = "Neutral"  # Workaround to enable colors in embedded IPy.

        embed(config=config,
              header='Welcome to the Windows Sandbox!\nUse the `sandbox` object to control the sandbox.')


if __name__ == '__main__':
    main()



from .utils.sandbox_feature_state import verify_sandbox_feature_is_enabled
from .folder_mapper import PythonMapper, PythonUserSitePackagesMapper
from .sandbox import OnlineSandbox, OfflineSandbox
from .config import SandboxConfig
import sys

# Do not allow importing if the feature isn't enabled
if 'pytest' not in sys.modules:
    verify_sandbox_feature_is_enabled()

_DEFAULT_FOLDER_MAPPERS = [PythonMapper(), PythonUserSitePackagesMapper()]


def new_sandbox(folder_mappers=None, networking=True, logon_script="", virtual_gpu=True):
    """
    Create a new sandbox or connect to an existing running instance.
    When networking=False:
    - Boots up a windows sandbox with the requested configuration.

    When networking=True:
    - Boots a sandbox with python mapped to it
    - Starts an RPyC server
    - Returns an OnlineSandbox instance with a `.rpyc` property.
    """

    folder_mappers = folder_mappers or []
    folder_mappers.extend(_DEFAULT_FOLDER_MAPPERS)

    config = SandboxConfig(folder_mappers=folder_mappers,
                           networking=networking,
                           logon_script=logon_script,
                           virtual_gpu=virtual_gpu)
    if networking:
        return OnlineSandbox(config)
    else:
        return OfflineSandbox(config)


def connect_to_sandbox():
    """
    Connect to an existing running sandbox which was created with `new_sandbox(networking=True)`.
    It connects to the already running RPyC server in the sandbox.
    """

    config = SandboxConfig(folder_mappers=[],
                           networking=True,
                           logon_script="",
                           virtual_gpu=True)
    return OnlineSandbox(config, launch_new_instance=False)



from ..utils.file import wait_for_file_creation
from ..utils.path import shared_folder_path_in_sandbox
from ..utils import dev_environment
from ..utils import venv

from ..folder_mapper import PythonMapper, FolderMapper, PythonUserSitePackagesMapper
from .. import target

import tempfile
import pathlib
import socket
import os
import sys


class ServerNotResponding(Exception):
    pass


def _is_server_responding(address, port):
    try:
        socket.create_connection((address, port), timeout=3).close()
        return True
    except socket.error:
        return False


def _get_shared_directory():
    shared_directory = pathlib.Path(tempfile.gettempdir()) / 'shared_windows_sandbox_dir'
    if not shared_directory.exists():
        shared_directory.mkdir()

    return shared_directory


class OnlineSession:
    def __init__(self, sandbox):
        self.sandbox = sandbox
        self.shared_directory = _get_shared_directory()
        self.server_address_path = pathlib.Path(self.shared_directory) / 'server_address'
        self.server_address_path_in_sandbox = shared_folder_path_in_sandbox(self.shared_directory) / 'server_address'

    def _get_logon_script(self, server_address_path):
        commands = []

        if dev_environment.is_dev_environment():
            # If we're in a dev environment we create the intermediate directory up to the egglink,
            # and then symlink the egglink to the mapped shared folder path.

            if dev_environment.get_egglink_path().drive != 'C:':
                # Handle this bizarre case by mapping a virtual drive from the actual drive letter to C:.
                # We'll then symlink the real development path to the shared folder on the desktop.
                commands.append('subst {} C:\\'.format(dev_environment.get_egglink_path().drive))

            commands.append('mkdir {}'.format(dev_environment.get_egglink_path().parent))

            commands.append('mklink /D {} {}'.format(
                dev_environment.get_egglink_path(),
                shared_folder_path_in_sandbox(dev_environment.get_egglink_path())))

        if venv.is_inside_venv():
            # If we're running in a virtual-environment we need to create a symlink between the
            # mapped base_prefix (shared folder path) and the path that the venv expects
            # the Python installation to be in.
            commands.append('mklink /D "{}" "{}"'.format(
                sys.base_prefix,
                shared_folder_path_in_sandbox(sys.base_prefix)))

        local_target_script_path = target.__file__
        remote_target_script_path = local_target_script_path.replace(str(PythonMapper().path()), str(
            shared_folder_path_in_sandbox(PythonMapper().path())))

        remote_python_path = sys.executable.replace(str(PythonMapper().path()), str(
            shared_folder_path_in_sandbox(PythonMapper().path())))

        user_site_path = PythonUserSitePackagesMapper().path()
        if user_site_path.exists():
            remote_user_site_packages_path = shared_folder_path_in_sandbox(user_site_path)
        else:
            remote_user_site_packages_path = ''

        # Launch the target script.
        commands.append(r'{} {} --disable-firewall {} {}'.format(
            remote_python_path,
            remote_target_script_path,
            str(server_address_path),
            remote_user_site_packages_path))
        print(commands[-1])

        return 'cmd.exe /c "{}"'.format(' && '.join(commands))

    def running_sandbox_server_information(self, allow_new_instance=False):
        """
        Get currently running server information (connection tuple).
        """
        try:
            # Try to connect to an already available targets.
            return self.connect_to_sandbox(timeout=0)
        except (ServerNotResponding, FileNotFoundError):
            # Server doesn't respond or a the connection data file doesn't exist.
            # We'll remove the connection data file and proceed.
            try:
                os.remove(str(self.server_address_path))
            except FileNotFoundError:
                pass

            if not allow_new_instance:
                raise ServerNotResponding("Sandbox is not currently running.")

    def configure_sandbox(self):
        """
        Configures the sandbox for a new online session.
        """

        extra_folder_mappers = [FolderMapper(self.shared_directory, read_only=False)]
        if dev_environment.is_dev_environment():
            # Let's map the egg link to the sandbox if we're in a dev environment.
            extra_folder_mappers.append(FolderMapper(dev_environment.get_egglink_path()))

        if venv.is_inside_venv():
            extra_folder_mappers.append(FolderMapper(sys.base_prefix))

        self.sandbox.config.logon_script = self._get_logon_script(self.server_address_path_in_sandbox)
        self.sandbox.config.folder_mappers.extend(extra_folder_mappers)

    def connect_to_sandbox(self, timeout=60):
        """
        Connect to the sandbox sever.
        Waits for the creation of the shared server address file, and checks that it responds.

        :returns: Connection tuple.
        """

        if wait_for_file_creation(str(self.server_address_path), timeout=timeout):
            address, port = self.server_address_path.read_text().split(':')
        else:
            raise FileNotFoundError("Sandbox server didn't startup")

        if not _is_server_responding(address, int(port)):
            raise ServerNotResponding()

        return address, int(port)



"""
This script is part of a Windows Shell Extension that allows you to run executable files
in a sandboxed environment using the right click menu.
Note that the script maps the entire parent directory of the executable,
this is done in order to preserve potential local dependencies.
"""

import sys
import argparse
import pathlib

import winsandbox


def read_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("executable", help="The executable path to run sandboxed")
    parser.add_argument("--network", default=False, action="store_true", help="Enable networking in the sandbox")
    parser.add_argument("--run", default=False, action="store_true", help="Run the executable inside the sandbox")

    return parser.parse_args()


def main():
    args = read_args()

    executable = pathlib.PurePath(args.executable)

    # Map executable folder to sandbox
    mapper = winsandbox.FolderMapper(folder_path=executable.parent, read_only=True)
    sandbox_mapped_dir = pathlib.PurePath(r'C:\Users\WDAGUtilityAccount\Desktop') / executable.parent.name

    if args.run:
        # Run the executable
        command = 'cmd /c "{}"'.format(sandbox_mapped_dir / executable.name)
    else:
        # Only open mapped folder
        command = 'explorer "{}"'.format(sandbox_mapped_dir)

    # Launch the sandbox
    winsandbox.new_sandbox(networking=args.network, logon_script=command, folder_mappers=[mapper])


if __name__ == '__main__':
    main()



import wmi

from .errors import SandboxFeatureIsNotAvailableError, SandboxFeatureIsNotEnabledError

_WINDOWS_SANDBOX_FEATURE_NAME = 'Containers-DisposableClientVM'
_ENABLED_STATE = 1


def is_sandbox_feature_enabled():
    """
    Checks whether or not the optional feature of windows sandboxes is enabled.
    """

    wmi_client = wmi.WMI()
    matching_features = wmi_client.Win32_OptionalFeature(Name=_WINDOWS_SANDBOX_FEATURE_NAME)
    if not matching_features:
        raise SandboxFeatureIsNotAvailableError()

    assert len(matching_features) == 1  # Make sure that there isn't any funny business and there's only 1 feature.
    sandbox_feature = matching_features[0]
    return sandbox_feature.InstallState == _ENABLED_STATE


def verify_sandbox_feature_is_enabled():
    if not is_sandbox_feature_enabled():
        raise SandboxFeatureIsNotEnabledError()



from codecs import open as codecs_open
from setuptools import setup, find_packages

# Get the long description from the relevant file
with codecs_open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(name='pywinsandbox',
      version='1.4.0',
      description=u"Python Utilities for Windows Sandbox",
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[],
      keywords='',
      author=u"Yiftach Karkason",
      author_email='ykarkason@gmail.com',
      url='https://github.com/karkason/pywinsandbox',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'yattag',
          'rpyc',
          'cached-property',
          'recordclass',
          'IPython',
          'wmi'
      ],
      extras_require={
          'test': ['pytest'],
      },
      entry_points="""
      [console_scripts]
      winsandbox=winsandbox.cli:main
      wsb=winsandbox.cli:main
      """
      )



import sys
import pathlib


def get_egglink_path():
    egg_link = pathlib.Path(sys.executable).parent / 'Lib' / 'site-packages' / 'pywinsandbox.egg-link'
    if egg_link.exists():
        return pathlib.Path(egg_link.read_text().splitlines()[0])

    return None


def is_dev_environment():
    return get_egglink_path() is not None



import yattag


def _get_boolean_text(value):
    return 'Default' if value else 'Disabled'


def _format_folder_mappers(folder_mappers, tag, text):
    with tag('MappedFolders'):
        for folder_mapper in folder_mappers:
            with tag('MappedFolder'):
                with tag('HostFolder'):
                    text(str(folder_mapper.path()))
                with tag('ReadOnly'):
                    text(str(folder_mapper.read_only()).lower())


def generate_config_file(config):
    config.folder_mappers = list(filter(lambda m: m.path().exists(), config.folder_mappers))

    document, tag, text = yattag.Doc().tagtext()

    with tag('Configuration'):
        with tag('VGpu'):
            text(_get_boolean_text(config.virtual_gpu))

        with tag('Networking'):
            text(_get_boolean_text(config.networking))

        with tag('LogonCommand'):
            with tag('Command'):
                text(config.logon_script)

        _format_folder_mappers(config.folder_mappers, tag, text)

    return yattag.indent(document.getvalue())



_WINDOWS_SANDBOX_ARTICLE_LINK = "https://techcommunity.microsoft.com/t5/windows-kernel-internals/windows-sandbox" \
                                "/ba-p/301849"


class SandboxFeatureIsNotAvailableError(RuntimeError):
    """
    Raised when failed getting any information about the windows sandbox feature.
    """
    def __init__(self):
        super(SandboxFeatureIsNotAvailableError, self).__init__(
            'Could not find the windows sandbox feature on this computer.\n'
            'Please visit "{}"\n'
            'for prerequisites for using this feature.'.format(_WINDOWS_SANDBOX_ARTICLE_LINK))


class SandboxFeatureIsNotEnabledError(RuntimeError):
    def __init__(self):
        """
        Raised when the windows sandbox is available on the computer but isn't enabled.
        """
        super(SandboxFeatureIsNotEnabledError, self).__init__(
            'The optional feature of windows sandbox is not enabled!\n'
            'Please visit "{}"\n'
            'for information on how to enable it.'.format(_WINDOWS_SANDBOX_ARTICLE_LINK))



import tempfile
import pathlib
from ..config import generate_config_file

WINDOWS_SANDBOX_CONFIG_FILE_SUFFIX = '.wsb'


class OfflineSession:
    def __init__(self, sandbox):
        self.sandbox = sandbox

    def run(self):
        config_file = generate_config_file(self.sandbox.config)

        # And start the sandbox.
        with tempfile.NamedTemporaryFile() as temp:
            config_file_path = pathlib.Path(temp.name).with_suffix(WINDOWS_SANDBOX_CONFIG_FILE_SUFFIX)
            config_file_path.write_text(config_file)
            self.sandbox.start_sandbox(str(config_file_path))
